{-#  LANGUAGE NPlusKPatterns  #-}

{-- 
    Primitive recursive functions in Haskell 
    Tuples of natural numbers are represented by lists

    Original version: WS-99-00 Prof. Dr. Raul Rojas
    modified SS-2010 by: Prof. Dr. Margarita Esponda
    modified WS-12-13 by: Prof. Dr. Margarita Esponda
--}

---------------------- Basic functions ---------------------------------

type PRFunction = ([Integer]->Integer)

-- zero function

z :: PRFunction
z xs = 0

--succesor function

s :: PRFunction
s [x] = x+1

--projection functions

p::Integer -> [Integer] -> Integer
p 1 (a:b) = a
p n (a:b) = p (n-1) b

---------------------- Metha-functions --------------------------------

-- for the composition scheme

compose :: PRFunction -> [PRFunction] -> [Integer] -> Integer
compose h gs xs = h [g xs | g<-gs ]

-- for the primitive recursive scheme

pr :: PRFunction -> PRFunction -> PRFunction -> [Integer] -> Integer

pr rec g h (  0  :xs) = g xs
pr rec g h ((n+1):xs) = h ( (rec (n:xs)):n:xs )

--------------------------------------------------------------------------
--------------------------------------------------------------------------

----------- lecture examples of primitive recursive functions ------------

-- predecessor
 
predd :: PRFunction
predd  = pr predd (const 0) (p 2) 

-- addition

add :: PRFunction
add  = pr add (p 1) (compose s [(p 1)]) 

-- test of equality to zero

isZero :: PRFunction
isZero  = pr isZero (const 1) (const 0) 

-- test of positive

isPositive :: PRFunction
isPositive = pr isPositive (const 0) (const 1)

-- smaller or equal (<=)

leq :: PRFunction
leq = compose isZero [sub]

-- smaller (<) 

smaller :: PRFunction
smaller = compose isPositive [compose sub [(p 2), (p 1)]]

--  greater (>)

greater = compose isPositive [compose sub [(p 1), (p 2)]]

-- greater or equal (<=)

geq :: PRFunction
geq = compose isZero [compose sub [(p 2), (p 1)]]

-- multiplication

mul :: PRFunction
mul  = pr mul z (compose add [(p 1),(p 3)]) 

-- subtraction 

sub :: PRFunction
sub  = compose sub' [(p 2),(p 1)] 
       where
         sub'  = pr sub' (p 1) (compose predd [(p 1)]) 

-- logical operators

nott :: PRFunction
nott  = isZero 

andd :: PRFunction
andd  = mul 

-- min function

minn = compose sub [ p 1, compose sub [p 1, p 2] ]

-- True if n is odd

odd2 [n] = pr odd2 (const 0) (compose nott [(p 1)]) [n]

-- True if n is even

even2 [n] = nott[odd2 [n]]

-- divide with 2

half [n] = pr half (const 0) (compose add [(p 1),(compose odd2 [(p 2)])]) [n]

-- factorial function

fact [n] = pr fact (const 1) (compose mul [(p 1),(compose s [(p 2)])]) [n]

-- integer division

idiv = compose predd [compose idiv' [compose add [(p 1), (const 1)], (p 2), (p 1)]]

idiv' [n,x,y] = pr idiv' g h [n,x,y]
               where
               g = const 0
               h = compose add [(p 1), compose nott [ compose sub [ compose mul [(p 2),(p 3)],(p 4)]]]

-- If then else

iff :: PRFunction
iff = pr iff (p 2) (p 3)

-- conditional (if-then-else)

cond :: PRFunction
cond = compose add [ fst, snd ]
        where
             fst  =  compose mul [ (compose sub [ (const 1), (compose isZero [p 1]) ]), (p 2) ]
             snd  =  compose mul [ (compose isZero [p 1]), (p 3)]

-- the same Conditional function but with explicit specification of the arguments

cond2 :: PRFunction
cond2 [b, first, second] = add [mul [sub [1, isZero [b]], first], mul [isZero [b], second]]